/***************************************************************************
 $RCSfile: ddvcard.c,v $
                             -------------------
    cvs         : $Id: ddvcard.c,v 1.14 2005/01/30 13:05:13 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#ifdef HAVE_I18N
# ifdef HAVE_LOCALE_H
#  include <locale.h>
# endif
# ifdef HAVE_LIBINTL_H
#  include <libintl.h>
# endif

# define I18N(msg) dgettext(PACKAGE, msg)
#else
# define I18N(msg) msg
#endif

#define I18N_NOOP(msg) msg


#include "ddvcard_p.h"

#include <chipcard2-client/cards/processorcard.h>

#include <aqbanking/banking.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/misc.h>
#include <gwenhywfar/padd.h>
#include <gwenhywfar/md.h>
#include <gwenhywfar/waitcallback.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>



GWEN_INHERIT(AH_MEDIUM, AH_MEDIUM_DDVCARD)
GWEN_INHERIT(AH_MEDIUMPROVIDER, AH_MEDIUMPROVIDER_DDVCARD)


AH_MEDIUM_DDVCARD *AH_MediumDDVCard_Data_new(){
  AH_MEDIUM_DDVCARD *mddv;

  GWEN_NEW_OBJECT(AH_MEDIUM_DDVCARD, mddv);
  mddv->localSignSeq=1;
  return mddv;
}



void AH_MediumDDVCard_Data_free(AH_MEDIUM_DDVCARD *mddv){
  if (mddv) {
    GWEN_DB_Group_free(mddv->dbContexts);
    GWEN_KeySpec_free(mddv->cryptKeySpec);
    GWEN_KeySpec_free(mddv->signKeySpec);
    free(mddv);
  }
}



AH_MEDIUM *AH_MediumDDVCard_new(AH_HBCI *hbci,
                                AH_MEDIUMPROVIDER *mp,
                                const char *mediumName) {
  AH_MEDIUM *m;
  AH_MEDIUM_DDVCARD *mddv;

  m=AH_Medium_new(hbci,
		  AH_MediumTypeDDV,
		  AH_MEDIUM_DDVCARD_NAME,
		  mediumName);
  mddv=AH_MediumDDVCard_Data_new();
  GWEN_INHERIT_SETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD,
		       m, mddv,
		       AH_MediumDDVCard_FreeData);
  mddv->pluginProvider=mp;

  AH_Medium_SetDeviceType(m, AH_MediumDeviceChipcard);

  /* set implemented functions for Medium */
  AH_Medium_SetMountFn(m, AH_MediumDDVCard_Mount);
  AH_Medium_SetCreateFn(m, AH_MediumDDVCard_Create);
  AH_Medium_SetUnmountFn(m, AH_MediumDDVCard_Unmount);
  AH_Medium_SetGetNextSignSeqFn(m, AH_MediumDDVCard_GetNextSignSeq);
  AH_Medium_SetSetLocalSignSeqFn(m, AH_MediumDDVCard_SetLocalSignSeq);
  AH_Medium_SetSignFn(m, AH_MediumDDVCard_Sign);
  AH_Medium_SetVerifyFn(m, AH_MediumDDVCard_Verify);
  AH_Medium_SetEncryptKeyFn(m, AH_MediumDDVCard_EncryptKey);
  AH_Medium_SetDecryptKeyFn(m, AH_MediumDDVCard_DecryptKey);
  AH_Medium_SetGenerateMsgKeyFn(m, AH_MediumDDVCard_GenerateMsgKey);
  AH_Medium_SetSelectContextFn(m, AH_MediumDDVCard_SelectContext);
  AH_Medium_SetCreateContextFn(m, AH_MediumDDVCard_CreateContext);

  AH_Medium_SetReadContextFn(m, AH_MediumDDVCard_ReadContext);
  AH_Medium_SetWriteContextFn(m, AH_MediumDDVCard_WriteContext);

  AH_Medium_SetToDbFn(m, AH_MediumDDVCard_toDb);
  AH_Medium_SetFromDbFn(m, AH_MediumDDVCard_fromDb);

  return m;
}



void AH_MediumDDVCard_FreeData(void *bp, void *p){
  AH_MEDIUM *m;
  AH_MEDIUM_DDVCARD *mddv;

  assert(bp);
  assert(p);
  m=(AH_MEDIUM*)bp;
  mddv=(AH_MEDIUM_DDVCARD*)p;
  AH_MediumDDVCard_Data_free(mddv);
}



int AH_MediumDDVCard_Mount(AH_MEDIUM *m){
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;
  GWEN_DB_NODE *dbCardData;
  LC_CLIENT *client;
  LC_CARD *hcard;
  int first;
  int havepin;
  const char *currCardNumber;
  int herr;
  GWEN_BUFFER *idBuf;
  GWEN_TYPE_UINT32 bid;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  client=AH_MediumDDVCardProvider_GetClient(mddv->pluginProvider);
  assert(client);

  bid=AB_Banking_ShowBox(AH_Medium_GetBankingApi(m),
                         0,
                         I18N("Accessing Card"),
                         I18N("Mounting card, please wait..."));

  res=LC_Client_StartWait(client, 0, 0);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not send StartWait request");
    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
    return -1;
  }

  first=1;
  for (;;) {
    int timeout;

    // determine timeout value
    if (first)
      timeout=3;
    else
      timeout=5;

    hcard=LC_Client_WaitForNextCard(client, timeout);
    if (!hcard) {
      int mres;

      mres=AH_Medium_UI(m, AH_Medium_UIActionInsertMedium);
      if (mres) {
	DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
	LC_Client_StopWait(client);
	AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
	return -1;
      }
    }
    else {
      /* ok, we have a card, now check it */
      if (LC_DDVCard_ExtendCard(hcard)) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "DDV card not available, please check your setup");
        LC_Card_free(hcard);
        LC_Client_StopWait(client);
	AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
        return -1;
      }

      res=LC_Card_Open(hcard);
      if (res!=LC_Client_ResultOk) {
        LC_Card_free(hcard);
        DBG_NOTICE(AQHBCI_LOGDOMAIN, "Could not open card, maybe not a DDV card?");
      } /* if card not open */
      else {
        const char *mname;

        mname=AH_Medium_GetMediumName(m);
        dbCardData=LC_DDVCard_GetCardDataAsDb(hcard);
	assert(dbCardData);

        currCardNumber=GWEN_DB_GetCharValue(dbCardData,
                                            "cardNumber",
                                            0,
                                            0);
	if (!currCardNumber) {
	  DBG_ERROR(AQHBCI_LOGDOMAIN, "INTERNAL: No card number in card data.");
	  abort();
	}

        DBG_NOTICE(AQHBCI_LOGDOMAIN, "Card number: %s", currCardNumber);

        if (!mname) {
	  DBG_NOTICE(AQHBCI_LOGDOMAIN, "No medium name");
          AH_Medium_SetMediumName(m, currCardNumber);
          mname=AH_Medium_GetMediumName(m);
          break;
        }
        if (!*mname) {
          DBG_NOTICE(AQHBCI_LOGDOMAIN, "Empty medium name");
	  AH_Medium_SetMediumName(m, currCardNumber);
          mname=AH_Medium_GetMediumName(m);
          break;
        }

        if (strcasecmp(mname, currCardNumber)==0) {
          DBG_NOTICE(AQHBCI_LOGDOMAIN, "Card number equals");
          break;
        }

        LC_Card_Close(hcard);
	LC_Card_free(hcard);

        hcard=LC_Client_PeekNextCard(client);
        if (!hcard) {
	  int mres;
    
	  mres=AH_Medium_UI(m, AH_Medium_UIActionInsertCorrectMedium);
	  if (mres) {
	    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
	    LC_Client_StopWait(client);
	    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
	    return -1;
	  }
	} /* if there is no other card waiting */
        else {
          /* otherwise there already is another card in another reader,
           * so no need to bother the user. This allows to insert all
           * cards in all readers and let me choose the card ;-) */
        } /* if there is another card waiting */
      } /* if card open */
    } /* if there is a card */

    first=0;
  } /* for */

  /* ok, now we have the card we wanted to have, now ask for the pin */
  LC_Client_StopWait(client);
  havepin=0;
  while(!havepin) {
    /* enter pin */
    herr=AH_MediumDDVCard__enterPin(m, hcard);
    if (herr) {
      LC_Card_Close(hcard);
      LC_Card_free(hcard);
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in PIN input");
      AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
      return -1;
    }
    else
      havepin=1;
  } /* while !havepin */

  /* ok, now we have a card and the PIN is verified, that's it ;-) */
  mddv->card=hcard;

  /* set security id */
  idBuf=LC_DDVCard_GetCardDataAsBuffer(hcard);
  assert(idBuf);
  AH_Medium_SetSecurityId(m,
                          GWEN_Buffer_GetStart(idBuf),
                          GWEN_Buffer_GetUsedBytes(idBuf));

  if (!AH_Medium_GetMediumName(m) && currCardNumber) {
    AH_Medium_SetMediumName(m, currCardNumber);
  }

  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Medium mounted");
  AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);

  if (mddv->signKeySpec==0 || mddv->cryptKeySpec==0 || !mddv->dbContexts) {
    bid=AB_Banking_ShowBox(AH_Medium_GetBankingApi(m),
			   0,
                           I18N("Accessing Card"),
                           I18N("Reading card data, please wait..."));
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			   0,
			   AB_Banking_LogLevelNotice,
			   I18N("First mount, will cache some data, "
				"please wait"));

    /* read keys if necessary */
    if (mddv->signKeySpec==0)
      herr=AH_MediumDDVCard__readSignKeySpecs(m);
    if (!herr && mddv->cryptKeySpec==0)
      herr=AH_MediumDDVCard__readCryptKeySpecs(m);
    if (herr) {
      mddv->card=0;
      LC_Card_Close(hcard);
      LC_Card_free(hcard);
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error reading key data");
      AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
      return -1;
    }

    /* read contexts */
    if (!mddv->dbContexts) {
      GWEN_TYPE_UINT32 bid;
      GWEN_DB_NODE *dbContexts;

      dbContexts=GWEN_DB_Group_new("contexts");
      res=LC_DDVCard_ReadInstituteData(mddv->card, 0, dbContexts);
      AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
      if (res!=LC_Client_ResultOk) {
	DBG_ERROR(AQHBCI_LOGDOMAIN, "No context available");
	GWEN_DB_Group_free(dbContexts);
      }
      else {
	mddv->dbContexts=dbContexts;
      }
    }
    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
  }

  return 0;
}



int AH_MediumDDVCard__enterPin(AH_MEDIUM *m, LC_CARD *hcard){
  LC_CLIENT_RESULT res;

  if (LC_Card_GetReaderFlags(hcard) & LC_CARD_READERFLAGS_KEYPAD) {
    int mres;

    DBG_INFO(AQHBCI_LOGDOMAIN,"Terminal has a keypad, will ask for pin.");
    /* tell the user about pin verification */
    mres=AH_Medium_UI(m, AH_Medium_UIActionStartInputPinViaKeypad);
    if (mres) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
      return -1;
    }
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                           0,
                           AB_Banking_LogLevelNotice,
                           I18N("Secure PIN input"));
    res=LC_DDVCard_SecureVerifyPin(hcard);
    /* tell the user about end of pin verification */
    mres=AH_Medium_UI(m, AH_Medium_UIActionFinishedInputPinViaKeypad);
    if (mres) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
      return -1;
    }

    if (res!=LC_Client_ResultOk) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                LC_Card_GetLastSW1(hcard),
                LC_Card_GetLastSW2(hcard),
                LC_Card_GetLastText(hcard));

      if (LC_Card_GetLastSW1(hcard)==0x63) {
	switch (LC_Card_GetLastSW2(hcard)) {
        case 0xc0:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelEmergency,
                                 I18N("Bad PIN, no error left"));
          AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                I18N("Serious Warning"),
                                I18N("The PIN you entered is invalid, there is "
                                "no room left for errors.\n"
                                "It is strongly recommended to "
                                "temporarily\n"
                                "stop using this current program.\n"
                                 "Please use the last program which "
                                 "successfully verified the PIN.\n"
                                 "This will reset the card's error "
                                 "counter.\n"
                                 "After that you may come back to this "
                                 "program."
                                 "<html>"
                                 "<p>"
                                 "The PIN you entered is invalid, there is "
                                 "no room left for errors."
                                 "</p>"
                                 "<p>"
                                 "It is <b>strongly</b> recommended to "
                                 "temporarily stop using this current "
                                 "program."
                                 "</p>"
                                 "<p>"
                                 "Please use the last program which "
                                 "successfully verified the PIN."
                                 "</p>"
                                 "<p>"
                                 "This will reset the card's error "
                                 "counter."
                                 "</p>"
                                 "<p>"
                                 "After that you may come back to this "
                                 "program."
                                 "</p>"
                                 "</html>"
                                 ),
                                I18N("Ok"), 0, 0);
          break;
        case 0xc1:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN, one error left"));
          break;
        case 0xc2:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN, two errors left"));
          break;
        default:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN"));
          break;
        } // switch
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
               LC_Card_GetLastSW2(hcard)==0x83) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Card unusable"));
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x64 &&
               LC_Card_GetLastSW2(hcard)==0x01) {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Aborted by user"));
        return -1;
      }
      else {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Unknown error"));
        return -1;
      }
    } /* if not ok */
    else {
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                             0,
                             AB_Banking_LogLevelNotice,
                             I18N("PIN ok."));
    }
  } /* if hasKeyPad */
  else {
    char pinBuffer[64];

    DBG_INFO(AQHBCI_LOGDOMAIN, "No keypad (or disabled), will ask for PIN");
    memset(pinBuffer, 0, sizeof(pinBuffer));
    if (AH_Medium_InputPin(m,
			   pinBuffer,
			   AH_MEDIUM_DDVCARD_PINMINLENGTH,
			   sizeof(pinBuffer),
			   AB_BANKING_INPUT_FLAGS_NUMERIC)) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error asking for PIN, aborting");
      memset(pinBuffer, 0, sizeof(pinBuffer));
      return -1;
    }

    DBG_INFO(AQHBCI_LOGDOMAIN, "Verifying the PIN");

    res=LC_DDVCard_VerifyPin(hcard, pinBuffer);
    if (res!=LC_Client_ResultOk) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                LC_Card_GetLastSW1(hcard),
                LC_Card_GetLastSW2(hcard),
                LC_Card_GetLastText(hcard));

      if (LC_Card_GetLastSW1(hcard)==0x63) {
	AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
	switch (LC_Card_GetLastSW2(hcard)) {
        case 0xc0:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelEmergency,
                                 I18N("Bad PIN, no error left"));
          AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                I18N("Serious Warning"),
                                I18N("The PIN you entered is invalid, there is "
                                "no room left for errors.\n"
                                "It is strongly recommended to "
                                "temporarily\n"
                                "stop using this current program.\n"
                                 "Please use the last program which "
                                 "successfully verified the PIN.\n"
                                 "This will reset the card's error "
                                 "counter.\n"
                                 "After that you may come back to this "
                                 "program."
                                 "<html>"
                                 "<p>"
                                 "The PIN you entered is invalid, there is "
                                 "no room left for errors."
                                 "</p>"
                                 "<p>"
                                 "It is <b>strongly</b> recommended to "
                                 "temporarily stop using this current "
                                 "program."
                                 "</p>"
                                 "<p>"
                                 "Please use the last program which "
                                 "successfully verified the PIN."
                                 "</p>"
                                 "<p>"
                                 "This will reset the card's error "
                                 "counter."
                                 "</p>"
                                 "<p>"
                                 "After that you may come back to this "
                                 "program."
                                 "</p>"
                                 "</html>"
                                 ),
                                I18N("Ok"), 0, 0);
          break;
        case 0xc1:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN, one error left"));
          break;
        case 0xc2:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN, two errors left"));
          break;
        default:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 I18N("Bad PIN"));
          break;
        } // switch
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
               LC_Card_GetLastSW2(hcard)==0x83) {
	AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Card unusable"));
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x64 &&
               LC_Card_GetLastSW2(hcard)==0x01) {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Aborted by user"));
        return -1;
      }
      else {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Unknown error");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               I18N("Unknown error"));
        return -1;
      }
    } // if not ok
    else {
      DBG_INFO(AQHBCI_LOGDOMAIN, "PIN ok");
      AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusOk);
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                             0,
                             AB_Banking_LogLevelNotice,
                             I18N("PIN ok."));
    }
  } // if no keyPad

  return 0;
}



int AH_MediumDDVCard__readSignKeySpecs(AH_MEDIUM *m){
  GWEN_KEYSPEC *ks;
  AH_MEDIUM_DDVCARD *mddv;
  int i;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  /* we do not set the ownerNames here, since we do not know it yet
   * It will be set as soon as a user is selected */
  ks=GWEN_KeySpec_new();
  GWEN_KeySpec_SetKeyType(ks, "des");
  GWEN_KeySpec_SetKeyName(ks, "S");
  i=LC_DDVCard_GetSignKeyNumber(mddv->card);
  if (i==-1) {
    DBG_WARN(AQHBCI_LOGDOMAIN,
	     "Could not get the sign key number, assuming 1");
    i=1;
  }
  GWEN_KeySpec_SetNumber(ks, i);

  i=LC_DDVCard_GetSignKeyVersion(mddv->card);
  if (i==-1) {
    DBG_WARN(AQHBCI_LOGDOMAIN,
	     "Could not get the sign key version, assuming 1");
    i=1;
  }
  GWEN_KeySpec_SetVersion(ks, i);

  AH_Medium_SetLocalSignKeySpec(m, ks);
  AH_Medium_SetRemoteSignKeySpec(m, ks);

  GWEN_KeySpec_free(mddv->signKeySpec);
  mddv->signKeySpec=ks;

  return 0;
}



int AH_MediumDDVCard__readCryptKeySpecs(AH_MEDIUM *m){
  GWEN_KEYSPEC *ks;
  AH_MEDIUM_DDVCARD *mddv;
  int i;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  ks=GWEN_KeySpec_new();
  GWEN_KeySpec_SetKeyType(ks, "des");
  GWEN_KeySpec_SetKeyName(ks, "V");
  i=LC_DDVCard_GetCryptKeyNumber(mddv->card);
  if (i==-1) {
    DBG_WARN(AQHBCI_LOGDOMAIN,
	     "Could not get the crypt key number, assuming 1");
    i=1;
  }
  GWEN_KeySpec_SetNumber(ks, i);

  i=LC_DDVCard_GetCryptKeyVersion(mddv->card);
  if (i==-1) {
    DBG_WARN(AQHBCI_LOGDOMAIN,
	     "Could not get the crypt key version, assuming 1");
    i=1;
  }
  GWEN_KeySpec_SetVersion(ks, i);

  AH_Medium_SetLocalCryptKeySpec(m, ks);
  AH_Medium_SetRemoteCryptKeySpec(m, ks);

  GWEN_KeySpec_free(mddv->cryptKeySpec);
  mddv->cryptKeySpec=ks;

  return 0;
}



int AH_MediumDDVCard__readKeys(AH_MEDIUM *m){
  AH_MEDIUM_DDVCARD *mddv;
  int rv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  rv=AH_MediumDDVCard__readSignKeySpecs(m);
  if (rv)
    return rv;
  rv=AH_MediumDDVCard__readCryptKeySpecs(m);
  if (rv)
    return rv;

  return 0;
}



int AH_MediumDDVCard_Create(AH_MEDIUM *m){
  AH_MEDIUM_DDVCARD *mddv;
  int rv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  rv=AH_MediumDDVCard_Mount(m);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  return 0;
}



int AH_MediumDDVCard_Unmount(AH_MEDIUM *m, int force){
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  mddv->selected=0;

  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Closing card");
  res=LC_Card_Close(mddv->card);
  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Deleting card");
  LC_Card_free(mddv->card);
  mddv->card=0;
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not unmount chip card");
    return -1;
  }
  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Medium unmounted");
  return 0;
}



AH_MEDIUM_RESULT AH_MediumDDVCard_Sign(AH_MEDIUM *m,
				       GWEN_BUFFER *msgbuf,
				       GWEN_BUFFER *signbuf){
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *hashbuf;
  char buffer[20];
  unsigned int bsize;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hash data");
  bsize=sizeof(buffer);
  if (GWEN_MD_Hash("RMD160",
                   GWEN_Buffer_GetStart(msgbuf),
                   GWEN_Buffer_GetUsedBytes(msgbuf),
                   buffer,
                   &bsize)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error hashing message");
    return AH_MediumResultGenericError;
  }
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hashing done");

  hashbuf=GWEN_Buffer_new(0, bsize, 0, 1);
  GWEN_Buffer_AppendBytes(hashbuf, buffer, bsize);

  /* sign hash */
  GWEN_Buffer_Rewind(hashbuf);
  res=LC_DDVCard_SignHash(mddv->card, hashbuf, signbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error signing hash (%d)", res);
    GWEN_Buffer_free(hashbuf);
    return AH_MediumResultGenericError;
  }

  GWEN_Buffer_free(hashbuf);
  return AH_MediumResultOk;
}



int AH_MediumDDVCard_GetNextSignSeq(AH_MEDIUM *m){
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *mbuf;
  GWEN_DB_NODE *dbRecord;
  int seq;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  res=LC_ProcessorCard_SelectEF(mddv->card, "EF_SEQ");
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return 0;
  }

  mbuf=GWEN_Buffer_new(0, 4, 0, 1);
  res=LC_ProcessorCard_ReadRecord(mddv->card, 1, mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(mbuf);
    return 0;
  }
  GWEN_Buffer_Rewind(mbuf);
  dbRecord=GWEN_DB_Group_new("seq");
  if (LC_Card_ParseRecord(mddv->card, 1, mbuf, dbRecord)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error parsing record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }

  seq=GWEN_DB_GetIntValue(dbRecord, "seq", 0, -1);
  if (seq==-1) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad record data in EF_SEQ");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }

  seq++;
  GWEN_DB_SetIntValue(dbRecord, GWEN_DB_FLAGS_OVERWRITE_VARS,
		      "seq", seq);
  GWEN_Buffer_Reset(mbuf);
  if (LC_Card_CreateRecord(mddv->card, 1, mbuf, dbRecord)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error creating record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }
  GWEN_Buffer_Rewind(mbuf);
  res=LC_ProcessorCard_WriteRecord(mddv->card, 1, mbuf);
  GWEN_DB_Group_free(dbRecord);
  GWEN_Buffer_free(mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return 0;
  }

  return seq;
}



int AH_MediumDDVCard_SetLocalSignSeq(AH_MEDIUM *m, int i) {
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *mbuf;
  GWEN_DB_NODE *dbRecord;
  int seq;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  res=LC_ProcessorCard_SelectEF(mddv->card, "EF_SEQ");
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return -1;
  }

  mbuf=GWEN_Buffer_new(0, 4, 0, 1);
  res=LC_ProcessorCard_ReadRecord(mddv->card, 1, mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(mbuf);
    return -1;
  }
  GWEN_Buffer_Rewind(mbuf);
  dbRecord=GWEN_DB_Group_new("seq");
  if (LC_Card_ParseRecord(mddv->card, 1, mbuf, dbRecord)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error parsing record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return -1;
  }

  seq=GWEN_DB_GetIntValue(dbRecord, "seq", 0, -1);
  if (seq==-1) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad record data in EF_SEQ");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return -1;
  }

  GWEN_DB_SetIntValue(dbRecord, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "seq", i);
  GWEN_Buffer_Reset(mbuf);
  if (LC_Card_CreateRecord(mddv->card, 1, mbuf, dbRecord)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error creating record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return -1;
  }
  GWEN_Buffer_Rewind(mbuf);
  res=LC_ProcessorCard_WriteRecord(mddv->card, 1, mbuf);
  GWEN_DB_Group_free(dbRecord);
  GWEN_Buffer_free(mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return -1;
  }

  return 0;
}



AH_MEDIUM_RESULT AH_MediumDDVCard_Verify(AH_MEDIUM *m,
					 GWEN_BUFFER *msgbuf,
					 GWEN_BUFFER *signbuf,
					 int signseq){
  AH_MEDIUM_DDVCARD *mddv;
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *hashbuf;
  GWEN_BUFFER *tmpsignbuf;
  char buffer[20];
  unsigned int bsize;
  unsigned int i;
  const char *p1, *p2;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hash data");
  bsize=sizeof(buffer);
  if (GWEN_MD_Hash("RMD160",
                   GWEN_Buffer_GetStart(msgbuf),
                   GWEN_Buffer_GetUsedBytes(msgbuf),
                   buffer,
                   &bsize)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error hashing message");
    return AH_MediumResultGenericError;
  }
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hashing done");

  hashbuf=GWEN_Buffer_new(0, bsize, 0, 1);
  GWEN_Buffer_AppendBytes(hashbuf, buffer, bsize);

  /* padd */
  //DBG_DEBUG(AQHBCI_LOGDOMAIN, "Padding hash using ISO 9796");
  //if (GWEN_Padd_PaddWithISO9796(hashbuf)) {
  //  DBG_INFO(AQHBCI_LOGDOMAIN, "here");
  //  GWEN_Buffer_free(hashbuf);
  //  return AH_MediumResultGenericError;
  //}

  /* sign hash */
  GWEN_Buffer_Rewind(hashbuf);
  tmpsignbuf=GWEN_Buffer_new(0, GWEN_Buffer_GetUsedBytes(signbuf), 0, 1);
  res=LC_DDVCard_SignHash(mddv->card, hashbuf, tmpsignbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error signing hash (%d)", res);
    GWEN_Buffer_free(hashbuf);
    GWEN_Buffer_free(tmpsignbuf);
    return AH_MediumResultGenericError;
  }

  p1=GWEN_Buffer_GetStart(signbuf);
  p2=GWEN_Buffer_GetStart(tmpsignbuf);
  bsize=GWEN_Buffer_GetUsedBytes(signbuf);
  if (bsize!=GWEN_Buffer_GetUsedBytes(tmpsignbuf)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Invalid signature (1) [%d!=%d]",
	      bsize, GWEN_Buffer_GetUsedBytes(tmpsignbuf));
    GWEN_Buffer_free(hashbuf);
    GWEN_Buffer_free(tmpsignbuf);
    return AH_MediumResultInvalidSignature;
  }

  /* compare signatures */
  for (i=0; i<bsize; i++) {
    if (*p1++!=*p2++) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Invalid signature (2)");
      GWEN_Buffer_free(hashbuf);
      GWEN_Buffer_free(tmpsignbuf);
      return AH_MediumResultInvalidSignature;
    }
  } /* for */

  GWEN_Buffer_free(hashbuf);
  GWEN_Buffer_free(tmpsignbuf);
  DBG_INFO(AQHBCI_LOGDOMAIN, "Signature is valid");
  return AH_MediumResultOk;
}



GWEN_BUFFER *AH_MediumDDVCard_GenerateMsgKey(AH_MEDIUM *m) {
  GWEN_BUFFER *kbuf;
  LC_CLIENT_RESULT res;
  AH_MEDIUM_DDVCARD *mddv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  kbuf=GWEN_Buffer_new(0, 16, 0, 1);

  /* create left */
  res=LC_DDVCard_GetChallenge(mddv->card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting random-1 (%d)", res);
    GWEN_Buffer_free(kbuf);
    return 0;
  }

  /* create right */
  res=LC_DDVCard_GetChallenge(mddv->card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting random-2 (%d)", res);
    GWEN_Buffer_free(kbuf);
    return 0;
  }

  return kbuf;
}



AH_MEDIUM_RESULT AH_MediumDDVCard_EncryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *encKey){
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *ibuf;
  AH_MEDIUM_DDVCARD *mddv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  /* check size */
  if (GWEN_Buffer_GetUsedBytes(srckey)!=16) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad key length (expected 16, got %d)",
              GWEN_Buffer_GetUsedBytes(srckey));
    return AH_MediumResultBadKey;
  }

  ibuf=GWEN_Buffer_new(0, 8, 0, 1);

  /* encode left */
  GWEN_Buffer_AppendBytes(ibuf,
                          GWEN_Buffer_GetStart(srckey), 8);
  res=LC_DDVCard_CryptBlock(mddv->card, ibuf, encKey);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error encrypting block (%d)", res);
    GWEN_Buffer_free(ibuf);
    return AH_MediumResultGenericError;
  }

  /* encode right */
  GWEN_Buffer_Reset(ibuf);
  GWEN_Buffer_AppendBytes(ibuf,
                          GWEN_Buffer_GetStart(srckey)+8, 8);
  res=LC_DDVCard_CryptBlock(mddv->card, ibuf, encKey);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error encrypting block (%d)", res);
    GWEN_Buffer_free(ibuf);
    return AH_MediumResultGenericError;
  }

  return AH_MediumResultOk;
}



AH_MEDIUM_RESULT AH_MediumDDVCard_DecryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *dstkey){
  AH_MEDIUM_RESULT rv;

  rv=AH_MediumDDVCard_EncryptKey(m, srckey, dstkey);
  if (rv!=AH_MediumResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not decrypt key");
  }
  return rv;
}



int AH_MediumDDVCard_SelectContext(AH_MEDIUM *m, int idx) {
  AH_MEDIUM_DDVCARD *mddv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    mddv->selected=0;
    return -1;
  }

  if (idx<0 || idx>5) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "Context index %d out of boundaries",
	      idx);
    return -1;
  }

  return 0;
}



int AH_MediumDDVCard_CreateContext(AH_MEDIUM *m,
				   int country,
				   const char *bankId,
				   const char *userId){
  AH_MEDIUM_DDVCARD *mddv;
  GWEN_DB_NODE *dbCurr;
  GWEN_DB_NODE *dbFreeContext;
  int freeIdx;
  int i;
  LC_CLIENT_RESULT res;
  GWEN_TYPE_UINT32 bid;
  int lcountry;
  const char *lbankId;
  const char *luserId;

  assert(m);
  assert(bankId);
  assert(userId);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mddv->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  freeIdx=-1;
  dbFreeContext=0;
  i=0;
  dbCurr=GWEN_DB_FindFirstGroup(mddv->dbContexts, "context");
  while(dbCurr) {
    lcountry=GWEN_DB_GetIntValue(dbCurr, "country", 0, 280);
    lbankId=GWEN_DB_GetCharValue(dbCurr, "bankCode", 0, 0);
    luserId=GWEN_DB_GetCharValue(dbCurr, "userId", 0, 0);
    if ((lcountry==0 || lbankId==0) && freeIdx==-1) {
      freeIdx=i;
      dbFreeContext=dbCurr;
    }
    if ((lcountry && country && lcountry==country) &&
	(lbankId && bankId && strcasecmp(lbankId, bankId)==0) &&
	(luserId && userId && strcasecmp(luserId, userId)==0)) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "Found context \"%d/%s/%s\"",
	       lcountry, lbankId, luserId);
      break;
    }
    dbCurr=GWEN_DB_FindNextGroup(dbCurr, "context");
    i++;
  } /* while */

  if (dbCurr) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context \"%d/%s/%s\" already exists",
	      country, bankId, userId);
    return -1;
  }

  if (freeIdx==-1) {
    if (i>5) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "All contexts in use");
      return -1;
    }
    dbFreeContext=GWEN_DB_GetGroup(mddv->dbContexts,
				   GWEN_PATH_FLAGS_CREATE_GROUP,
				   "context");
  }
  assert(dbFreeContext);
  GWEN_DB_SetIntValue(dbFreeContext, GWEN_DB_FLAGS_OVERWRITE_VARS,
		      "country", country);
  GWEN_DB_SetCharValue(dbFreeContext, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "bankCode", bankId);
  GWEN_DB_SetCharValue(dbFreeContext, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "userId", userId);

  bid=AB_Banking_ShowBox(AH_Medium_GetBankingApi(m),
			 0,
                         I18N("Accessing Card"),
                         I18N("Writing card, please wait..."));
  res=LC_DDVCard_WriteInstituteData(mddv->card, freeIdx+1, dbFreeContext);
  AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error writing context %d", freeIdx);
    return -1;
  }
  return freeIdx;
}



int AH_MediumDDVCard_ReadContext(AH_MEDIUM *m,
				 int idx,
				 int *country,
				 GWEN_BUFFER *bankId,
				 GWEN_BUFFER *userId,
				 GWEN_BUFFER *server,
				 int *port){
  AH_MEDIUM_DDVCARD *mddv;
  GWEN_DB_NODE *dbT;
  const char *s;
  int i;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mddv->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  i=0;
  dbT=GWEN_DB_FindFirstGroup(mddv->dbContexts, "context");
  while(dbT) {
    if (i==idx)
      break;
    dbT=GWEN_DB_FindNextGroup(dbT, "context");
    i++;
  } /* while */

  if (!dbT) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context %d not found", idx);
    return -1;
  }

  if (country)
    *country=GWEN_DB_GetIntValue(dbT, "country", 0, 280);
  if (bankId) {
    s=GWEN_DB_GetCharValue(dbT, "bankCode", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(bankId, s);
  }

  if (userId) {
    s=GWEN_DB_GetCharValue(dbT, "userId", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(userId, s);
  }

  if (server) {
    s=GWEN_DB_GetCharValue(dbT, "comAddress", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(server, s);
  }

  if (port) {
    i=GWEN_DB_GetIntValue(dbT, "comService", 0, -1);
    if (i==2)
      *port=3000;
    else if (i==3)
      *port=443;
  }

  return 0;
}



int AH_MediumDDVCard_WriteContext(AH_MEDIUM *m,
				  int idx,
				  int country,
				  const char *bankId,
				  const char *userId,
				  const char *server,
				  int port){
  AH_MEDIUM_DDVCARD *mddv;
  GWEN_DB_NODE *dbT;
  int i;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mddv->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  i=0;
  dbT=GWEN_DB_GetFirstGroup(mddv->dbContexts);
  while(dbT) {
    if (strcasecmp(GWEN_DB_GroupName(dbT), "context")==0) {
      if (i==idx)
        break;
    }
    dbT=GWEN_DB_GetNextGroup(dbT);
    i++;
  } /* while */

  if (!dbT) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context %d not found", idx);
    return -1;
  }

  if (country)
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"country", country);
  if (bankId)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "bankCode", bankId);
  if (userId)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "userId", userId);
  if (server)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "comAddress", server);
  if (port==3000)
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 2);
  else if (port==443) {
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 3);
  }
  else if (port) {
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 0);
  }

  return 0;
}



int AH_MediumDDVCard_toDb(const AH_MEDIUM *m, GWEN_DB_NODE *db){
  GWEN_DB_NODE *dbT;
  AH_MEDIUM_DDVCARD *mddv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  if (mddv->dbContexts) {
    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "contexts");
    assert(dbT);
    GWEN_DB_AddGroupChildren(dbT, mddv->dbContexts);
  }

  if (mddv->signKeySpec) {
    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "signKeySpec");
    assert(dbT);
    GWEN_KeySpec_ToDb(mddv->signKeySpec, dbT);
  }

  if (mddv->cryptKeySpec) {
    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "cryptKeySpec");
    assert(dbT);
    GWEN_KeySpec_ToDb(mddv->cryptKeySpec, dbT);
  }

  return 0;
}



int AH_MediumDDVCard_fromDb(AH_MEDIUM *m, GWEN_DB_NODE *db){
  GWEN_DB_NODE *dbT;
  AH_MEDIUM_DDVCARD *mddv;

  assert(m);
  mddv=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_DDVCARD, m);
  assert(mddv);

  GWEN_KeySpec_free(mddv->cryptKeySpec); mddv->cryptKeySpec=0;
  GWEN_KeySpec_free(mddv->signKeySpec);  mddv->signKeySpec=0;

  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "contexts");
  if (dbT) {
    GWEN_DB_Group_free(mddv->dbContexts);
    mddv->dbContexts=GWEN_DB_Group_dup(dbT);
  }

  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "signKeySpec");
  if (dbT) {
    mddv->signKeySpec=GWEN_KeySpec_new();
    if (GWEN_KeySpec_FromDb(mddv->signKeySpec, dbT)) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad signKeySpec");
      GWEN_KeySpec_free(mddv->signKeySpec);  mddv->signKeySpec=0;
    }
    else {
      AH_Medium_SetLocalSignKeySpec(m, mddv->signKeySpec);
      AH_Medium_SetRemoteSignKeySpec(m, mddv->signKeySpec);
    }
  }

  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "cryptKeySpec");
  if (dbT) {
    mddv->cryptKeySpec=GWEN_KeySpec_new();
    if (GWEN_KeySpec_FromDb(mddv->cryptKeySpec, dbT)) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad cryptKeySpec");
      GWEN_KeySpec_free(mddv->cryptKeySpec); mddv->cryptKeySpec=0;
    }
    else {
      AH_Medium_SetLocalCryptKeySpec(m, mddv->cryptKeySpec);
      AH_Medium_SetRemoteCryptKeySpec(m, mddv->cryptKeySpec);
    }
  }

  return 0;
}














AH_MEDIUMPROVIDER *ddvcard_MediumProvider_new(AH_HBCI *hbci){
  AH_MEDIUMPROVIDER *mp;
  AH_MEDIUMPROVIDER_DDVCARD *mpddv;
#ifdef HAVE_I18N
  const char *s;
#endif

#ifdef HAVE_I18N
  setlocale(LC_ALL,"");
  s=bindtextdomain(PACKAGE,  LOCALEDIR);
  if (s) {
    DBG_NOTICE(AQHBCI_LOGDOMAIN, "Locale bound.");
    bind_textdomain_codeset(PACKAGE, "UTF-8");
  }
  else {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error binding locale");
  }
#endif

  mp=AH_MediumProvider_new(hbci, AH_MEDIUM_DDVCARD_NAME);
  AH_MediumProvider_SetDeviceType(mp, AH_MediumDeviceChipcard);
  AH_MediumProvider_SetFactoryFn(mp, AH_MediumDDVCardProvider_Factory);
  AH_MediumProvider_SetCheckFn(mp, AH_MediumDDVCardProvider_Check);

  GWEN_NEW_OBJECT(AH_MEDIUMPROVIDER_DDVCARD, mpddv);
  mpddv->client=LC_Client_new("AqHBCI_DDVCard", VERSION, 0);
  GWEN_INHERIT_SETDATA(AH_MEDIUMPROVIDER, AH_MEDIUMPROVIDER_DDVCARD,
                       mp, mpddv,
                       AH_MediumDDVCardProvider_freeData);

  if (LC_Client_ReadConfigFile(mpddv->client, 0)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error reading configuration.\n");
    AH_MediumProvider_free(mp);
    return 0;
  }

  return mp;
}



void AH_MediumDDVCardProvider_freeData(void *bp, void *p) {
  AH_MEDIUMPROVIDER_DDVCARD *mpddv;

  mpddv=(AH_MEDIUMPROVIDER_DDVCARD*)p;
  LC_Client_free(mpddv->client);
  GWEN_FREE_OBJECT(mpddv);
}



AH_MEDIUM*
AH_MediumDDVCardProvider_Factory(AH_MEDIUMPROVIDER *mp,
				 AH_HBCI *hbci,
				 const char *mediumName) {
  return AH_MediumDDVCard_new(hbci,
                              mp,
			      mediumName);
}



AH_MEDIUM_CHECKRESULT
AH_MediumDDVCardProvider_Check(AH_MEDIUMPROVIDER *mp,
			       AH_HBCI *hbci,
                               GWEN_BUFFER *mediumName) {
  AH_MEDIUMPROVIDER_DDVCARD *mpddv;
  LC_CARD *card;
  LC_CLIENT_RESULT res;
  GWEN_DB_NODE *dbCardData;
  const char *p;
  GWEN_TYPE_UINT32 bid;

  assert(mp);
  mpddv=GWEN_INHERIT_GETDATA(AH_MEDIUMPROVIDER,AH_MEDIUMPROVIDER_DDVCARD, mp);
  assert(mpddv);

  bid=AB_Banking_ShowBox(AH_HBCI_GetBankingApi(hbci),
                         0,
                         I18N("Accessing Card"),
                         I18N("Checking medium, please wait..."));

  res=LC_Client_StartWait(mpddv->client, 0, 0);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Wait request timed out");
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }

  card=LC_Client_WaitForNextCard(mpddv->client, 30000);
  if (!card) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "ERROR: No card found");
    LC_Client_StopWait(mpddv->client);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }
  LC_Client_StopWait(mpddv->client);

  if (LC_DDVCard_ExtendCard(card)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "ERROR: Could not extend card as DDV card");
    LC_Card_free(card);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }

  res=LC_Card_Open(card);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN,
	     "Probably not a DDV card (or IO error, check logs)");
    LC_Card_Close(card);
    LC_Card_free(card);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultWrongMedium;
  }

  dbCardData=LC_DDVCard_GetCardDataAsDb(card);
  assert(dbCardData);
  p=GWEN_DB_GetCharValue(dbCardData,
                         "cardNumber",
                         0,
                         0);
  if (!p) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "INTERNAL: No card number in card data.");
    abort();
  }

  if (GWEN_Buffer_GetUsedBytes(mediumName)==0) {
    /* set name */
    GWEN_Buffer_AppendString(mediumName, p);
  }
  else {
    /* compare name */
    if (strcasecmp(GWEN_Buffer_GetStart(mediumName), p)!=0) {
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Wrong medium name");
      LC_Card_Close(card);
      LC_Card_free(card);
      AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
      return AH_MediumCheckResultWrongName;
    }
  }

  LC_Card_Close(card);
  LC_Card_free(card);

  /* do nothing for now */
  AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
  return AH_MediumCheckResultOk;
}



LC_CLIENT *AH_MediumDDVCardProvider_GetClient(const AH_MEDIUMPROVIDER *mp){
  AH_MEDIUMPROVIDER_DDVCARD *mpddv;

  assert(mp);
  mpddv=GWEN_INHERIT_GETDATA(AH_MEDIUMPROVIDER,
                             AH_MEDIUMPROVIDER_DDVCARD,
                             mp);
  return mpddv->client;
}



















