/***************************************************************************
 $RCSfile: ddvcard_p.h,v $
                             -------------------
    cvs         : $Id: ddvcard_p.h,v 1.4 2005/01/30 13:05:13 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/


#ifndef AH_MEDIUM_DDVCARD_P_H
#define AH_MEDIUM_DDVCARD_P_H

#define AH_MEDIUM_DDVCARD_NAME "DDVCard"
#define AH_MEDIUM_DDVCARD_VMAJOR 1
#define AH_MEDIUM_DDVCARD_VMINOR 0


#define AH_MEDIUM_DDVCARD_PINMINLENGTH 4

#include <chipcard2-client/cards/ddvcard.h>
#include <chipcard2-client/client/client.h>

#include "ddvcard.h"


typedef struct AH_MEDIUM_DDVCARD AH_MEDIUM_DDVCARD;
struct AH_MEDIUM_DDVCARD {
  int selected;

  GWEN_TYPE_UINT32 localSignSeq;
  GWEN_TYPE_UINT32 remoteSignSeq;

  LC_CARD *card;
  AH_MEDIUMPROVIDER *pluginProvider;

  GWEN_DB_NODE *dbContexts;

  GWEN_KEYSPEC *signKeySpec;
  GWEN_KEYSPEC *cryptKeySpec;
};


AH_MEDIUM_DDVCARD *AH_MediumDDVCard_Data_new();
void AH_MediumDDVCard_Data_free(AH_MEDIUM_DDVCARD *mrdh);

int AH_MediumDDVCard_GetNextSignSeq(AH_MEDIUM *m);
int AH_MediumDDVCard_SetLocalSignSeq(AH_MEDIUM *m, int i);

int AH_MediumDDVCard_Mount(AH_MEDIUM *m);
int AH_MediumDDVCard_Create(AH_MEDIUM *m);
int AH_MediumDDVCard_Unmount(AH_MEDIUM *m, int force);

AH_MEDIUM_RESULT AH_MediumDDVCard_Sign(AH_MEDIUM *m,
                                       GWEN_BUFFER *msgbuf,
                                       GWEN_BUFFER *signbuf);
int AH_MediumDDVCard_GetNextSignSeq(AH_MEDIUM *m);
AH_MEDIUM_RESULT AH_MediumDDVCard_Decrypt(AH_MEDIUM *m,
                                          GWEN_BUFFER *msgbuf,
                                          GWEN_BUFFER *decryptbuf,
                                          GWEN_BUFFER *msgKeyBuffer);
AH_MEDIUM_RESULT AH_MediumDDVCard_Verify(AH_MEDIUM *m,
                                         GWEN_BUFFER *msgbuf,
                                         GWEN_BUFFER *signbuf,
                                         int signseq);
AH_MEDIUM_RESULT AH_MediumDDVCard_Encrypt(AH_MEDIUM *m,
                                          GWEN_BUFFER *msgbuf,
                                          GWEN_BUFFER *encryptbuf,
                                          GWEN_BUFFER *msgKeyBuffer);

int AH_MediumDDVCard_SelectContext(AH_MEDIUM *m, int idx);

int AH_MediumDDVCard_CreateContext(AH_MEDIUM *m,
                                   int country,
                                   const char *bankId,
                                   const char *userId);

int AH_MediumDDVCard_ReadContext(AH_MEDIUM *m,
                                 int idx,
                                 int *country,
                                 GWEN_BUFFER *bankId,
                                 GWEN_BUFFER *userId,
                                 GWEN_BUFFER *server,
                                 int *port);

int AH_MediumDDVCard_WriteContext(AH_MEDIUM *m,
                                  int idx,
                                  int country,
                                  const char *bankId,
                                  const char *userId,
                                  const char *server,
                                  int port);


void AH_MediumDDVCard_FreeData(void *bp, void *p);


int AH_MediumDDVCard__enterPin(AH_MEDIUM *m, LC_CARD *hcard);

int AH_MediumDDVCard__readKeys(AH_MEDIUM *m);

GWEN_BUFFER *AH_MediumDDVCard_GenerateMsgKey(AH_MEDIUM *m);
AH_MEDIUM_RESULT AH_MediumDDVCard_EncryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *encKey);
AH_MEDIUM_RESULT AH_MediumDDVCard_DecryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *encKey);

int AH_MediumDDVCard_toDb(const AH_MEDIUM *m, GWEN_DB_NODE *db);
int AH_MediumDDVCard_fromDb(AH_MEDIUM *m, GWEN_DB_NODE *db);


int AH_MediumDDVCard__readSignKeySpecs(AH_MEDIUM *m);
int AH_MediumDDVCard__readCryptKeySpecs(AH_MEDIUM *m);







typedef struct AH_MEDIUMPROVIDER_DDVCARD AH_MEDIUMPROVIDER_DDVCARD;
struct AH_MEDIUMPROVIDER_DDVCARD {
  LC_CLIENT *client;
};

void AH_MediumDDVCardProvider_freeData(void *bp, void *p);

AH_MEDIUM*
  AH_MediumDDVCardProvider_Factory(AH_MEDIUMPROVIDER *mp,
                                   AH_HBCI *hbci,
                                   const char *mediumName);

AH_MEDIUM_CHECKRESULT
  AH_MediumDDVCardProvider_Check(AH_MEDIUMPROVIDER *mp,
                                 AH_HBCI *hbci,
                                 GWEN_BUFFER *mediumName);

LC_CLIENT *AH_MediumDDVCardProvider_GetClient(const AH_MEDIUMPROVIDER *mp);


#endif /* AH_MEDIUM_DDVCARD_H */







